=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed October 2012 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   TopoShaper_Tool.rb
# Original Date	:   20 Oct 12 - version 1.0
# Description	:   TopoShaper Interactive Tool
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module TopoShaper

T6[:TIT_Operation] = "TopoShaper: Terrain"
T6[:TIP_GridLimits] = "Grid limits: Cell number X[%1, %2] Y[%3, %4] - Cell dimensions X[%5, %6] Y[%7, %8]"
T6[:TIP_GridLengthEquiv] = "Cell Dimensions: X = %1 Y = %2"
T6[:BOX_Grid_Title] = "Grid"

T6[:BOX_Mode_selection] = "SELECTION"
T6[:BOX_Mode_cleansing] = "CLEANSING"
T6[:BOX_Mode_algo] = "PREVIEW"
T6[:BOX_Mode_geometry] = "GENERATION"
T6[:TIP_Mode_selection] = "Step 1: Selection of contours"
T6[:TIP_Mode_cleansing] = "Step 2: Cleansing of contours"
T6[:TIP_Mode_algo] = "Step 3: Calculation of terrain and Preview"
T6[:TIP_Mode_geometry] = "Step 4: Generation of terrain"

T6[:TIP_Rollback_ToSelection] = "Go back to Contours Selection"
T6[:TIP_Rollback_ToCleansing] = "Go back to Contour Cleansing"
T6[:TIP_Rollback_ToAlgo] = "Go back to Terrain Preview"

T6[:BOX_OptionGeometry] = "Elements"
T6[:TIP_OptionGeometry] = "Options for generating the terrain"
T6[:TIP_OptionGeometryWall] = "Include / Exclude the Skirt"
T6[:TIP_OptionGeometryContour] = "Include / Exclude the Contours"
T6[:TIP_OptionGeometryMap] = "Include / Exclude 2D Contour Map"

T6[:BOX_Hilltop] = "Hilltops"
T6[:TIP_HilltopMain] = "Global Option for extrapolating on hilltops and basins bottom"
T6[:TIP_HilltopFlatAll] = "Flat on ALL hilltops and basins"
T6[:TIP_HilltopFlat] = "Flat on selected hilltops and basins"
T6[:TIP_HilltopAutoAll] = "Rounding on ALL hilltops and basins"
T6[:TIP_HilltopAuto] = "Rounding on selected hilltops and basins"

T6[:BOX_Validate_selection] = "Cleanse | Contours"	
T6[:BOX_Validate_cleansing] = "Calculate | Terrain"	
T6[:BOX_Validate_algo] = "Generate | Terrain"	
T6[:TIP_Validate_selection] = "Validate selection and go to Contour Cleansing"	
T6[:TIP_Validate_cleansing] = "Go to the Calculation and Preview of the Terrain"	
T6[:TIP_Validate_algo] = "Generate the terrain as a SU Group"	
	
T6[:BOX_Cliff] = "Cliff"
T6[:TIP_CliffMain] = "Parameter to force cliff at contour (steeper slope)"
T6[:TIP_CliffNO] = "NO Cliff"
T6[:TIP_CliffVal] = "Steepness of the cliff (0 = No Cliff, 100% = Full Cliff)"
T6[:TIP_CliffTrue] = "TRUE Cliff"

T6[:BOX_WorkingView] = "Working View"	
T6[:TIP_WorkingView] = "Working view showing terrain and contours only"	
T6[:TIP_PreviewMode_InSitu] = "Native SU view in the model with focus on contours"	
T6[:TIP_PreviewMode_Surface] = "Working view centered around 3D Terrain"	
T6[:TIP_PreviewMode_Map] = "Working view centered around 2D Contour Map"	
T6[:TIP_PreviewMode_SurfaceMap] = "Working view showing both 3D Terrain and 2D Contour Map"	

T6[:BOX_SaveReplace] = "Update Contours & Reload"	
T6[:TIP_SaveReplace] = "Update contours by replacing existing ones, and Reload them"	
T6[:BOX_SaveCopy] = "Save Copy & Reload"	
T6[:TIP_SaveCopy] = "Save the updated contours as a copy in a group"	

T6[:TIP_ContourAccept] = "Include the contour(s) in the terrain generation"
T6[:TIP_ContourIgnore] = "Exclude the contour(s) from the terrain generation"

T6[:TIP_EditTerrain] = "Click to edit Terrain"
T6[:MNU_EditTerrain] = "Edit Terrain"

T6[:TIP_DebugOption] = "Show contours used by each node of the mesh"
T6[:ERROR_CannotLoadAttr] = "ERROR: Cannot load the Terrain"

T6[:OPS_UC_UpdateTitle] = "TOPOSHAPER: Update contours"
T6[:VBAR_UC_UpdateTitle] = "Updating Contours Geometry"
T6[:VBAR_UC_CheckDuplicate] = "Checking duplicated groups"
T6[:VBAR_UC_MakeUnique] = "Making groups Unique"
T6[:VBAR_UC_HandlingUnique] = "Handling Entities of Unique groups"
T6[:VBAR_UC_EraseNeedleEyes] = "Erasing Needle Eyes"
T6[:VBAR_UC_EraseEdges] = "Erasing Hooks and Modified edges"
T6[:VBAR_UC_CreateNewEdges] = "Generating the updated contours"
T6[:VBAR_UC_RegisterContours] = "Registering the contours"

T6[:MSG_RubyError] = "The operation is fully aborted"	
T6[:BOX_ClickToExit] = "Click to Exit"

#Structure to hold information on the original contours
Contour3D = Struct.new :inum, :pts, :lvx, :ledges, :tr, :parent

#--------------------------------------------------------------------------------------------------------------
# Top Calling functions: create the classes and launch the tools
#--------------------------------------------------------------------------------------------------------------			 				   

#Actual launching of menus	
def TopoShaper.action__mapping(action_code)
	case action_code
	when :quad_terrain, :cleanser
		Sketchup.active_model.select_tool TopoShaperTool.new(action_code)
	end	
end

#=============================================================================================
#=============================================================================================
# Class TopoShaperTool: main class for the Interactive tool
#=============================================================================================
#=============================================================================================

class TopoShaperTool < Traductor::PaletteSuperTool

@@dico = "TOPOSHAPER"
@@dico_grid = "Grid"

@@lst_options_global = [[:edgepicker_modifier, 'F'], [:edgepicker_anglemax, 40], [:edgepicker_stop_at_crossing, true],
                        [:edge_prop_filter, 'P'], [:geometry_wall, true], [:geometry_contour, false], [:geometry_map, false]]

def initialize(action_code, *args)
	#Loading parameters
	options_load_global

	#Basic initialization
	@model = Sketchup.active_model
	@tr_id = Geom::Transformation.new
	@title = "TopoShaper"
	@action_code = action_code
	
	#Parsing the arguments
	args.each { |arg| arg.each { |key, value|  parse_args(key, value) } if arg.class == Hash }

	#Static initialization
	init_cursors
	init_colors
	init_messages

	#Initializating the Edge Picker
	init_edge_picker
	
	#Creating the palette manager and texts
	init_palette
	
	init_VCB
	
	#Creating the Zoom artefact
	@zoomar = Traductor::ZoomArtefact.new
	
	@use_working_view = true
end

#TOPOSHAPER_TOOL: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /from/i
		@invoked_from = value
	end	
end	

#---------------------------------------------------------------------------------------------
# OPTIONS: Option Management and storage
#---------------------------------------------------------------------------------------------
				
#OPTIONS: Load options from registry
def options_load_global
	#Global GUI parameters for TopoShaper
	sparam = Sketchup.read_default "TopoShaper", "Options"
	hsh = {}
	if sparam
		hsh = eval(sparam) rescue nil
	end
	@@lst_options_global.each do |a|
		key, default = a
		hsh[key] = default unless hsh.has_key?(key)
	end
	@hsh_options_global = hsh
end

#OPTIONS: Save options to registry
def options_save
	#Getting the edgepicker parameters
	hsh = @selmode.get_hparams
	@hsh_options_global[:edgepicker_modifier] = hsh[:modifier]
	@hsh_options_global[:edgepicker_anglemax] = hsh[:anglemax]
	@hsh_options_global[:edgepicker_stop_at_crossing] = hsh[:stop_at_crossing]
	@hsh_options_global[:edgepicker_edge_prop_filter] = hsh[:edge_prop_filter]
	
	#Saving global options
	sparam = @hsh_options_global.inspect.gsub('"', "'")
	Sketchup.write_default "TopoShaper", "Options", sparam	
end
		
#INIT: Initialize texts and messages
def init_messages
	@mnu_exit = T6[:T_BUTTON_Exit]	
	@mnu_abort = T6[:T_STR_AbortTool]
	@tip_validate = @title	
	@tip_rollback_to_selection = T6[:TIP_Rollback_ToSelection]
	@tip_rollback_to_cleansing = T6[:TIP_Rollback_ToCleansing]
	@tip_rollback_to_algo = T6[:TIP_Rollback_ToAlgo]
	@mnu_validate_selection = T6[:TIP_Validate_selection]	
	@mnu_validate_cleansing = T6[:TIP_Validate_cleansing]
	@mnu_validate_algo = T6[:TIP_Validate_algo]	
	
	@tip_next_curve = T6[:TIP_NextCurve]
	@tip_exec_geometry = T6[:TIP_ExecGeometry]	
	@tip_edit_terrain = T6[:TIP_EditTerrain]
	
	@hsh_boxinfo_exit = { :bk_color => 'lightgreen', :fr_color => 'green' }					 	
end

#INIT: Initialize colors
def init_colors
	@color_message = 'khaki'
	@hsh_boxinfo_tip = { :bk_color => 'lightyellow', :fr_color => 'yellow' }					 
end

#INIT: Initialize cursors
def init_cursors
	@id_cursor_default = Traductor.create_cursor "Cursor_Arrow_Default", 2, 2
	@id_cursor_arrow_exit = Traductor.create_cursor "Cursor_Arrow_Exit", 0, 0	
	@id_cursor_validate = Traductor.create_cursor "Cursor_Validate", 0, 0	
	@id_cursor_hourglass_green = Traductor.create_cursor "Cursor_hourGlass_Green", 16, 16		
	@id_cursor_hourglass_red = Traductor.create_cursor "Cursor_hourGlass_Red", 16, 16		
	@id_cursor = @id_cursor_default
end

#INIT: Initializating the Edge Selection Picker
def init_edge_picker
	hsh = {}
	hsh[:notify_proc] = self.method 'notify_edge_picked'
	hsh[:modifier] = @hsh_options_global[:edgepicker_modifier]
	hsh[:anglemax] = @hsh_options_global[:edgepicker_anglemax]
	hsh[:stop_at_crossing] = @hsh_options_global[:edgepicker_stop_at_crossing]
	hsh[:edge_prop_filter] = @hsh_options_global[:edgepicker_edge_prop_filter]
	hsh[:title] = @title
	hsh[:no_vcb] = true
	hsh[:vbox_option_exit] = true
	@selmode = Traductor::EdgePicker.new hsh
end

#--------------------------------------------------
# Activation / Deactivation
#--------------------------------------------------

#ACTIVATION: Tool activation
def activate
	LibFredo6.register_ruby "TopoShaper"
	Traductor::Hilitor.stop_all_active
	@model = Sketchup.active_model
	@model_current = @model if @model != @model_current
	@view_tracker = G6::ViewTracker.new
	@selection = @model.selection
	@entities = @model.active_entities
	@view = @model.active_view
		
	#recovering problems in previous sessions
	recover_problems_with_previous_sessions
	
	#Resetting the environment
	reset_context
	
	#Initiating the palette
	set_state_mode :selection

	#Initializing the Executor of Operation
	hsh = { :title => T6[:TIT_Operation], :palette => @palette, :end_proc => self.method('geometry_terminate'),
            :no_commit => true }
	@suops = Traductor::SUOperation.new hsh
	@suops.start_operation

	#Handling the initial selection
	handle_initial_selection
		
	@view.invalidate
	show_message
end

#ACTIVATION: Recover problems with previous sessions
def recover_problems_with_previous_sessions
	#For Zoom artefacts
	Traductor::TempStuff.cleanup
	
	#for Views
	vdesc = @view.camera.description
	if vdesc && vdesc =~ /TOPOSHAPER/
		@view.camera.description = ""
	end	
end

#ACTIVATION: Reset context variables
def reset_context
	@button_down = false
	@ip = Sketchup::InputPoint.new
	@ph = @view.pick_helper
	@preview_in_situ = true
end

#ACTIVATION: Tool Deactivation
def deactivate(view)
	GridDimensionDialog.close
	#@myclass.edgewire_erase if @myclass
	#@zoomar.terminate
	restore_view
	@suops.finish_operation
	options_save
	selection_at_exit
	view.invalidate
end

#ACTIVATION: Manage the initial selection or the previous results
def handle_initial_selection
	@old_selection = @selection.to_a.clone
	return false if @selection.empty?
	
	#Selection of an existing terrain group
	lst_groups = @selection.grep(Sketchup::Group)
	lst_groups.each do |group|
		@type_attr = attribute_group?(group)
		if @type_attr
			@group_attr = group
			@tr_attr = group.transformation
			@initial_selection_ok = true
			algo_load_from_attr
			return
		end
	end
	
	#Selection by Edge picker
	@handling_initial_selection = true
	@selmode.check_initial_selection
	@lst_info_contours = @selmode.get_info_contours
	@handling_initial_selection = false
	return false if @lst_info_contours.length == 0
	@initial_selection_ok = true
	execute_validate
end
	
#ACTIVATION: Set / reset the selection at exit
def selection_at_exit
	sel = @old_selection.find_all { |a| a.valid? && a.instance_of?(Sketchup::Edge) } if @old_selection
	@selection.add sel unless sel.empty?
end
		
#ACTIVATION: Exit the tool
def exit_tool
	@aborting = false
	unless @mode == :geometry
		restore_view
		@suops.abort_operation
	end	
	@model.select_tool nil
end

#--------------------------------------------------
# Contextual menu
#--------------------------------------------------

#Contextual menu
def getMenu(menu)
	cxmenu = Traductor::ContextMenu.new
	
	if @mode == :algo
		@algo.contextual_menu_contribution(cxmenu)
	elsif @mode == :cleansing	
		@cleanser.contextual_menu_contribution(cxmenu)
	elsif @mode == :selection
		if @group_attr
			cxmenu.add_item(T6[:MNU_EditTerrain]) { algo_load_from_attr }
		else
			@selmode.contextual_menu_contribution(cxmenu)
		end	
	end
		
	#Menu Validate and Rollback
	cxmenu.add_sepa
	
	unless notify_from_palette(:gray, :validate) || (@group_attr && @mode == :selection)
		tip = notify_from_palette :tip, :validate
		cxmenu.add_item(tip) { notify_from_palette :exec, :validate }
	end
	unless notify_from_palette :gray, :back
		tip = notify_from_palette :tip, :back
		cxmenu.add_item(tip) { notify_from_palette :exec, :back }
	end

	#Exit
	cxmenu.add_sepa
	cxmenu.add_item(@mnu_abort) { abort_tool }

	#Exit
	if @mode == :geometry
		cxmenu.add_sepa
		cxmenu.add_item(@mnu_exit) { exit_tool }
	end
	
	#Showing the menu
	cxmenu.show menu
	true
end

#--------------------------------------------------
# State machine and Top Execution
#--------------------------------------------------

#EXEC: Set the current mode
def set_state_mode(mode)
	@mode = mode
	@palette.visible_family @mode
	@myclass = nil
	case @mode
	when :selection
		@mode_text = T6[:BOX_Mode_selection]
		@mode_tip = T6[:TIP_Mode_selection]	
	when :cleansing
		@algo.edgewire_erase if @algo
		@myclass = @cleanser
		@mode_text = T6[:BOX_Mode_cleansing]
		@mode_tip = T6[:TIP_Mode_cleansing]
	when :algo
		@cleanser.edgewire_erase if @cleanser
		@myclass = @algo
		@mode_text = T6[:BOX_Mode_algo]
		@mode_tip = T6[:TIP_Mode_algo]
	when :geometry
		@algo.edgewire_erase if @algo
		@mode_text = T6[:BOX_Mode_geometry]
		@mode_tip = T6[:TIP_Mode_geometry]	
	else
		@mode_text = ""
		@mode_tip = ""	
	end	
	onSetCursor
	@view.invalidate
end

#EXEC: Add or remove entity from selection
def notify_edge_picked(action, ledges)
	if (action == :outside || action == :finish) && @selmode.empty?
		exit_tool
	elsif action == :finish
		after_selection
	end	
	ledges
end

def ask_grid_dimension
	GridDimensionDialog.invoke @algo, self
end

def notify(action, *args)
	case action
	when :grid_dimension
		if args[0]
			@algo_processing = true
			@algo.notify_grid_dimension *args
			@algo_processing = false
			set_state_mode :algo
		end	
		
	when :infobox
		set_infobox *args
	end	
end

#Manage action after end of selection
def after_selection
	@lst_info_contours = @selmode.get_info_contours
	register_contours
	
	#Cleansing the contours
	cleansing_processing
end

#Rescue from Ruby errors and abort
def rescue_errors(exception, title)
	Traductor::Hilitor.stop_all_active
	Traductor::Hilitor.stop_all_active
	Traductor::RubyErrorDialog.invoke exception, title, T6[:MSG_RubyError, "TopoShaper"]
	abort_tool
end

#EXEC: Execute the Cleansing
def cleansing_processing
	@cleansing_processing = true
	begin
		lst_pts = (@selmode.empty?) ? @algo.get_lst_pts : @lst_pts
		@cleanser = TopoShaperCleanser.new(self, lst_pts)
		set_state_mode :cleansing
		status = @cleanser.top_processing @saved_contours
		set_preview_working(:map)
	rescue  Exception => e
		rescue_errors e, "CLEANSING"
	end	
	@cleansing_processing = false
	onMouseMove_zero
	onSetCursor
end

#EXEC: Execute the calculation of terrain
def algo_processing
	@algo_processing = true
	unless @cleanser
		hsh = { :palette => @palette, :group_attr => @group_attr, :tr_attr => @tr_attr }
	else
		hsh = { :palette => @palette, :plane_normal => Z_AXIS, :lst_pts => @cleanser.get_lst_pts }
	end	
	@algo = TopoShaperAlgo.new(self, @hsh_options_global, hsh)
	if @algo.creation_invalid?
		UI.messagebox T6[:ERROR_CannotLoadAttr]
		abort_tool
		return
	end
	@myclass.edgewire_erase if @myclass
	@myclass = @algo
	begin
		@algo.top_processing
		(@use_working_view) ? set_preview_working(:surface) : switch_view
	rescue Exception => e
		rescue_errors e, "CALCULATING TERRAINS"
	end	
	@algo_processing = false
	set_state_mode :algo
	@myclass.edgewire_show if working_view_active?
	onMouseMove_zero
	onSetCursor
end

#ATTR: Erase the attribute group if any
def attribute_group_retrieve
	return nil unless @entity_attr
	@model.active_entities.grep(Sketchup::Group).find { |g| g.entityID == @entity_attr }
end

#EXEC: Load a terrain from a group already constructed
def algo_load_from_attr
	@entity_attr = @group_attr.entityID
	@camera_last_working = nil
	algo_processing
end

#EXEC: Load a terrain from a group already cleansed
def cleansing_load_from_attr
	@entity_attr = nil
	@camera_last_working = nil
	cleansing_processing
end

#ROLLBACK: Check if rollback is allowed
def rollback_allowed?
	return false if @mode == :selection || (@mode == :cleansing && @group_attr)
	true
end

#ROLLBACK: Execute the rollback
def rollback_execute(no_undo=false)
	if @mode == :cleansing
		restore_view
		set_state_mode :selection
		@suops.abort_restart_operation
	elsif @mode == :algo
		if @cleanser
			set_state_mode :cleansing
			set_preview_working(:map) if @use_working_view	
			@myclass.edgewire_show if working_view_active?
			@myclass.compute_infobox
		else
			cleansing_processing
		end	
	elsif @mode == :geometry
		geometry_rollback
		set_state_mode :algo
		set_preview_working(:surface) if @use_working_view	
		@myclass.edgewire_show if working_view_active?
		@myclass.compute_infobox
	end	
	onMouseMove_zero
end

#Check if current contour can be validated
def validate_allowed?
	@group_attr || !(@selmode.empty?)
end

#Check if current contour can be validated
def validate_current_allowed?
	!@selmode.contours_validated?
end

#Validation of current curve and go to next
def validate_current_execute
	@selmode.validate_current_selection if validate_current_allowed?
end

#Validate action, depending on the current mode
def execute_validate
	case @mode
	when :selection
		@selmode.terminate_current_selection
	when :cleansing
		begin
			@cleanser.top_repairing
			set_preview_working(:surface)
		rescue Exception => e
			rescue_errors e, "REPAIRING CONTOURS"
		end
		algo_processing
	when :algo
		geometry_execute
	when :execute
		exit_tool
	end	
end

#Update the contour geometry
def save_copy_contours
	puts "Copy contours not yet implemented"
end
	
def save_replace_contours
	begin
		@cleanser.top_repairing
		status = geometry_update_contours
	rescue Exception => e
		rescue_errors e, "SAVE CONTOUR GEOMETRY"
	end
	if status
		@saved_contours = true
		cleansing_processing
	end	
end

#Abort the operation and exit
def abort_tool
	@aborting = true
	restore_view
	@suops.abort_operation
	@model.select_tool nil
end

#---------------------------------------------------------------------------------------------
# SU CONTOURS: Update contour sand reconstruct the 3D
#---------------------------------------------------------------------------------------------

#Update the geometry of contours
def geometry_update_contours
	return false unless @lst_info_contours
	
	#Getting the bad edges from the Edgepicker
	hbad_info = (@selmode) ? @selmode.get_bad_edges_info : []
	
	#Calculating the mapping between old and new contours
	lst_map = @cleanser.save_replace_contours
	return false if lst_map.nitems == 0

	#Starting the Visual Bar
	hsh = { :style_color => :bluegreen }
	vgbar = Traductor::VisualProgressBar.new T6[:VBAR_UC_UpdateTitle], hsh
	nbstep = 7
	istep = 0.0
	vgbar.start
	
	#Updating the contour geometry
	@suops.abort_operation
	@suops.start_operation T6[:OPS_UC_UpdateTitle]
	
	#Check if some parents are not unique and store edge transposition info
	vgbar.progression istep / nbstep, T6[:VBAR_UC_CheckDuplicate]
	hsh_not_uniq = {}
	hsh_mapid = {}
	n = @lst_info_contours.length
	n2 = 2 * n
	min = slice = 20
	mini = (n >= min && n >= 2*slice)
	@lst_info_contours.each_with_index do |info, ik| 
		vgbar.mini_progression(ik * 1.0 / n, 1.0 / nbstep) if mini && ik.modulo(slice) == 0
		lvx, ledges, tr, parent = info
		pid = parent.entityID
		next if hsh_not_uniq[pid] || G6.grouponent_unique?(parent)
		hsh_not_uniq[pid] = parent
		hsh_mapid[pid] = G6.grouponent_map_of_entities_object_id(parent)
	end
	
	#Make parent unique if required
	istep += 1
	vgbar.progression istep / nbstep, T6[:VBAR_UC_MakeUnique]
	hsh_not_uniq.each do |pid, parent|
		G6.grouponent_make_unique parent
	end

	#Reconstructing the list of edges and vertices when parents have been made unique
	istep += 1
	vgbar.progression istep / nbstep, T6[:VBAR_UC_HandlingUnique]
	@lst_info_contours.each_with_index do |info, ik| 
		vgbar.mini_progression((ik) * 1.0 / n, 1.0 / nbstep) if mini && (ik == 0 || (ik).modulo(slice) == 0)
		lvx, ledges, tr, parent = info
		pid = parent.entityID
		next unless hsh_not_uniq[pid]
		entities = G6.grouponent_entities(parent)
		hmap = hsh_mapid[pid]
		ledges = ledges.collect { |e| entities[hmap[e.object_id]] }
		lvx = G6.vertices_from_edges_in_sequence(ledges)
		@lst_info_contours[ik] = [lvx, ledges, tr, parent] 
	end
	
	#Erasing the bad edges
	istep += 1
	vgbar.progression istep / nbstep, T6[:VBAR_UC_EraseNeedleEyes]
	hbad_info.each do |pid, a|
		parent, lbad_edges = a
		entities = G6.grouponent_entities(parent)
		hmap = hsh_mapid[pid]
		lbad_edges = lbad_edges.collect { |e| entities[hmap[e.object_id]] } if hmap
		entities.erase_entities lbad_edges if lbad_edges.length > 0
	end
		
	#Erasing all required edges
	istep += 1
	vgbar.progression istep / nbstep, T6[:VBAR_UC_EraseEdges]
	new_info = []
	@lst_info_contours.each_with_index do |info, ik| 
		lvx, ledges, tr, parent = info
		entities = G6.grouponent_entities(parent)
		pts = lst_map[ik]		
		if pts == nil
			new_info.push info
		else
			ledges = ledges.find_all { |e| e.valid? }
			entities.erase_entities ledges
		end	
	end	
	
	#Creating the updated contours
	istep += 1
	vgbar.progression istep / nbstep, T6[:VBAR_UC_CreateNewEdges]
	@lst_info_contours.each_with_index do |info, ik| 
		lvx, ledges, tr, parent = info		
		pts = lst_map[ik]
		next if pts == :delete || pts == nil
		entities = G6.grouponent_entities(parent)
		t = tr.inverse
		edges = entities.add_curve pts.collect { |pt| t * pt }
		lvx = G6.vertices_from_edges_in_sequence(edges)
		info = [lvx, edges, tr, parent]
		new_info.push info
	end	
	@suops.finish_operation
	@suops.start_operation
	
	#Registering the contours
	istep += 1
	vgbar.progression istep / nbstep, T6[:VBAR_UC_RegisterContours]
	@lst_info_contours = new_info
	register_contours
	@selmode.reset_selection
	@lst_info_contours.each_with_index do |info, ik| 
		vgbar.mini_progression((ik) * 1.0 / n, 1.0 / nbstep) if mini && (ik == 0 || (ik).modulo(slice) == 0)
		lvx, ledges, tr, parent = info		
		@selmode.register_edges ledges, tr, parent
	end	
	
	#Stopping the visual bar
	vgbar.stop
	
	true
end

def save_attribute

end

#---------------------------------------------------------------------------------------------
# CONTOUR MANAGEMENT: 
#---------------------------------------------------------------------------------------------

#Creating a contour structure
def register_contours
	#@lst_contours_3d = []
	@lst_pts = []
	@lst_info_contours.each_with_index do |info, ik| 
		lvx, ledges, tr, parent = info
		#contour = Contour3D.new
		#contour.lvx, contour.ledges, contour.tr, contour.parent = info
		#contour.pts = lvx.collect { |vx| tr * vx.position }
		@lst_pts[ik] = lvx.collect { |vx| tr * vx.position }
		#contour.inum = ik
		#@lst_contours_3d.push contour
	end	
end

#---------------------------------------------------------------------------------------------
# SU TOOL: Mouse Movement Methods
#---------------------------------------------------------------------------------------------

#SU TOOL: Mouse Movements
def onMouseMove_zero(flag=nil) ; onMouseMove(@flags, @xmove, @ymove, @view) if @xmove ; end
def onMouseMove(flags, x, y, view)
	#Event for the palette
	@xmove = x
	@ymove = y
	if super
		@mouse_in_palette = true
		view.invalidate
		show_message
		return
	end	
	@mouse_in_palette = false
	#reconcile_button_state(flags, x, y, view)
	
	#Synchronize draw and move
	return if @moving
	@moving = true

	#Dispatching the event
	case @mode
	when :selection	
		@group_attr, @tr_attr, @type_attr = attribute_under_mouse(flags, x, y, view)
		@selmode.onMouseMove(flags, x, y, view) unless @group_attr
	when :algo, :cleansing
		if @myclass
			@myclass.mouse_on_move(flags, x, y, view)
			@zoomar.move_mark @myclass.zoom_target
		end	
	end
	
	#Refreshing the view
	view.invalidate
	show_message
end	

#SU TOOL: Check if the state of mouse is compatible with internal flags
def reconcile_button_state(flags, x, y, view)
	if RUN_ON_MAC
		onLButtonUp(flags, x, y, view) if flags == 256 && @button_down
	else
		onLButtonUp(flags, x, y, view) if flags == 1 && !@button_down
	end
end

def onMouseLeave(view)
	@mouseOut = true
	show_message
	view.invalidate
end

def onMouseEnter(view)
	@mouseOut = false
	view.invalidate
end
	
def attribute_group?(group)
	return nil unless group && group.attribute_dictionary("TOPOSHAPER", false)
	(group.get_attribute(@@dico, @@dico_grid) ? :algo : :cleanser)
end

def attribute_under_mouse(flags, x, y, view)
	@ph.do_pick x, y
	elt = [@ph.picked_face, @ph.picked_edge].find { |e| e }
				
	#Finding the parent and transformation
	tr = @tr_id
	parent = nil
	grand_parent = nil
	return nil unless elt
	for i in 0..@ph.count
		ls = @ph.path_at(i)
		if ls && ls.include?(elt)
			parent = ls[-2]
			grand_parent = ls[-3]
			tr = @ph.transformation_at(i)
			break
		end
	end	
	return nil unless parent && parent.instance_of?(Sketchup::Group)
	
	type_attr = attribute_group?(parent)
	return [parent, tr, type_attr] if type_attr
	
	type_attr = attribute_group?(grand_parent)
	return [grand_parent, parent.transformation * tr, type_attr] if type_attr

	nil
end
		
#---------------------------------------------------------------------------------------------
# SU TOOL: Click Management
#---------------------------------------------------------------------------------------------

#SU TOOL: Button click DOWN
def onLButtonDown(flags, x, y, view)
	#Interrupting geometry construction if running
	return if @suops.interrupt?

	#Palette management
	return if super
	
	#Dispatching the event
	case @mode
	when :selection
		@selmode.onLButtonDown(flags, x, y, view) unless @group_attr
	when :algo, :cleansing
		@myclass.onLButtonDown(flags, x, y, view)
		#execute_validate
	when :geometry
		exit_tool
	end	
	onMouseMove_zero
end

#SU TOOL: Button click UP - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	return if super
	case @mode
	when :selection	
		if @group_attr
			algo_load_from_attr
		else
			@selmode.onLButtonUp(flags, x, y, view)
		end	
	when :algo, :cleansing
		@myclass.onLButtonUp(flags, x, y, view)
	end	
	@button_down = false
	onMouseMove_zero
end

#SU TOOL: Double Click received
def onLButtonDoubleClick(flags, x, y, view)
	return if super
	case @mode
	when :selection	
		@selmode.onLButtonDoubleClick(flags, x, y, view)
	when :algo, :cleansing
		@myclass.onLButtonDoubleClick(flags, x, y, view)
	end	
	view.invalidate
end

#---------------------------------------------------------------------------------------------
# SU TOOL: Keyboard handling
#---------------------------------------------------------------------------------------------

#Return key pressed
def onReturn(view)
	execute_validate
end

#SU TOOL: Handle Key down
def onKeyDown(key, rpt, flags, view)
	key = Traductor.new_check_key key, flags, false
	
	#Interrupting geometry construction if running
	return if @suops.interrupt?
		
	@num_keydown = 0 unless @num_keydown
	@num_keydown += 1
	case key			
	when CONSTRAIN_MODIFIER_KEY
		@shift_down = true
		
	when COPY_MODIFIER_KEY
		@ctrl_down = @num_keydown
		onMouseMove_zero 1
			
	when 13
		if @shift_down || @ctrl_down
			
		end	
			
	else	
		@shift_down = @ctrl_down = false
	end	
end

#SU TOOL: Key UP received
def onKeyUp(key, rpt, flags, view)
	key = Traductor.new_check_key key, flags, true
	
	#Dispatching the event
	case key	
	when VK_DELETE, 8
		#geometry_rollback
		
	when CONSTRAIN_MODIFIER_KEY
		@time_shift_up = Time.now
		@shift_down = false
		
	when COPY_MODIFIER_KEY
		@time_ctrl_up = Time.now
		if @ctrl_down && @ctrl_down == @num_keydown

		end	
		onMouseMove_zero 0
		@ctrl_down = false
		
	else 
		if @time_ctrl_up && (Time.now - @time_ctrl_up) < 0.1	
			onMouseMove_zero 0
		end	
		if @time_shift_up && (Time.now - @time_shift_up) < 0.1	
			onMouseMove_zero 0
		end	
		@shift_down = @ctrl_down = false
	end	
end

#---------------------------------------------------------------------------------------------
# SU TOOL: Drawing Methods
#---------------------------------------------------------------------------------------------

#SU TOOL: Draw top method
def draw(view)	
	#Drawing the curves selected
	case @mode
	when :selection
		if @group_attr
			drawing_group_attr view
		else	
			@selmode.draw view unless @handling_initial_selection
		end	
	when :cleansing, :algo
		drawing_preview view
		@myclass.dessin_boxinfo(view, @xmove, @ymove) if @myclass && !@mouse_in_palette && !@mouseOut
	when :geometry
		G6.draw_rectangle_multi_text view, @xmove, @ymove, T6[:BOX_ClickToExit], @hsh_boxinfo_exit if !@mouse_in_palette && !@mouseOut
	end	
	@moving = false
	
	#Drawing the palette
	super
end

def drawing_preview(view)
	@myclass.dessin view	
end

#Draw the selection box for Terrain group
def drawing_group_attr(view)
	return unless @group_attr && @group_attr.valid?
	view.line_stipple = ''
	view.line_width = 3
	view.drawing_color = 'red'
	llines = G6.grouponent_box_lines(view, @group_attr, @tr_attr)
	view.draw GL_LINES, llines unless llines.empty?
	
	G6.draw_rectangle_multi_text view, @xmove, @ymove, @tip_edit_terrain, @hsh_boxinfo_tip
end

def getExtents
	(@myclass && @myclass.bb_extents) ? @myclass.bb_extents : @model.bounds
	@model.bounds
end

def suspend(view)
	if @mode == :algo || @mode == :cleansing
		@update_busy = true
		@myclass.edgewire_show if working_view_active?
	end
end

def resume(view)
	return view.invalidate unless @myclass && (@mode == :algo || @mode == :cleansing)
	if @mode == :algo
		@zoomar.move_mark @myclass.zoom_target
		@myclass.edgewire_hide if working_view_active?
	end	
	@myclass.dessin_refresh_when_view_changed
	update_dessin_on_resume
end

def update_dessin_on_resume
	if @update_busy
		@update_busy = false
		@update_timer = UI.start_timer(0.5) { update_dessin_on_resume } unless @update_timer
	else	
		@update_timer = nil
		@update_busy = false
		@myclass.dessin_update_when_view_changed_async
	end	
	@view.invalidate
end
#---------------------------------------------------------------------------------------------
# SU TOOL: Show Messages
#---------------------------------------------------------------------------------------------

#SU TOOL: Computing Current cursor
def onSetCursor
	#Hourglass cursors
	@id_cursor = 0
	if @algo_processing	|| @cleansing_processing
		@id_cursor = @id_cursor_hourglass_red
	elsif @geometry_processing
		@id_cursor = @id_cursor_hourglass_green
	end	
	return UI.set_cursor(@id_cursor) if @id_cursor != 0

	#Palette cursors
	ic = super
	return (ic != 0) if ic
	
	#Other cursors depending on state
	case @mode
	when :algo
		#@id_cursor = @id_cursor_validate
		@id_cursor = 0
	when :geometry
		@id_cursor = @id_cursor_arrow_exit
	else
		@id_cursor = @id_cursor_default
	end	
	UI.set_cursor @id_cursor
end

#SU TOOL: Show message in the palette
def show_message
	#Mouse if either in the palette or out of the viewport
	if @mouseOut
		palette_set_message nil
		return
	elsif @mouse_in_palette
		palette_set_message @palette.get_main_tooltip, 'aliceblue'
		return
	end	
	
	text = ""
	text = @myclass.palette_message if @myclass
	
	#Displaying the message
	palette_set_message text, @color_message
end

#SU TOOL: Display the message in the palette zone and SU status bar
def palette_set_message(msg, color=nil)
	@palette.set_message msg, color
	Sketchup.set_status_text((msg) ? msg : "")
	@view.invalidate
end

#SU TOOL: Display the message in the palette zone and SU status bar
def palette_set_tooltip(msg, level=nil)
	@palette.set_tooltip msg, level
end

#TOOL: Set the text for the information box in the palette
def set_infobox(text1, text2=nil, level=nil)
	@info_text1 = text1 if text1
	@info_text2 = text2 if text2
	text = @info_text1 + "\n" + @info_text2
	@palette.set_tooltip text, level
end

#--------------------------------------------------------------
#--------------------------------------------------------------
# Palette Management
#--------------------------------------------------------------
#--------------------------------------------------------------

#--------------------------------------------------
# Callback methods from palette
#--------------------------------------------------

#Notification method from palette for Back and Exec buttons
def notify_from_palette(action, code)
	case action
	when :tip
		compute_tooltip_for_palette code
	when :gray	
		compute_gray_for_palette code
	when :exec
		execute_from_palette code
	end	
end

#Tooltip for Back and Execute buttons
def compute_tooltip_for_palette(code)
	case code
	when :validate
		case @mode
		when :selection
			@mnu_validate_selection
		when :cleansing
			@mnu_validate_cleansing
		when :algo
			@mnu_validate_algo
		end	
	when :back
		case @mode
		when :cleansing
			@tip_rollback_to_selection
		when :algo
			@tip_rollback_to_cleansing
		when :geometry
			@tip_rollback_to_algo
		end	
	when :next_curve
		@tip_next_curve
	when :grid_nx_ny
		sl = @algo.grid_max_dimensions + @algo.grid_max_dimensions_delta.collect { |a| Sketchup.format_length a }
		T6[:TIP_GridLimits, *sl]
	when :grid_lx_ly
		sl = @algo.grid_current_delta.collect { |a| Sketchup.format_length a }
		T6[:TIP_GridLengthEquiv, *sl]
	else
		""
	end
end

#Gray status for Back and Execute buttons
def compute_gray_for_palette(code)
	case code
	when :validate
		!validate_allowed?		
	when :back
		!rollback_allowed?
	when :next_curve
		!validate_current_allowed?
	end	
end

#Execution for Back and Execute buttons
def execute_from_palette(code)
	case code
	when :validate
		execute_validate
	when :back
		rollback_execute
	when :next_curve
		validate_current_execute
	end	
end

#---------------------------------------------------------------------
# Creation of the palette
#---------------------------------------------------------------------

#PALETTE: Separator for the Main and floating palette
def pal_separator ;	@palette.declare_separator ; end

#PALETTE: Initialize the main palette and top level method
def init_palette
	hshpal = { :width_message => 400, :width_message_min => 150, :key_registry => :toposhaper }
	@palette = Traductor::Palette.new hshpal
	notify_proc = self.method 'notify_from_palette'
	@draw_local = self.method "draw_button_opengl"
	@blason_proc = self.method "draw_button_blason"
	@symb_blason = :blason
	@bk_color_tool = "lemonchiffon"
	@info_text1 = ""
	@info_text2 = ""
	@info_message = ""
		
	#Blason Button
	tip = "TopoShaper" + ' ' + T6[:T_STR_DefaultParamDialog]
	hsh = { :blason => true, :draw_proc => @blason_proc, :tooltip => tip }
	@palette.declare_button(@symb_blason, hsh) { MYPLUGIN.invoke_default_parameter_dialog }

	#Step Label
	palette_step_label
	
	#Modifiers for Selection mode
	@palette.set_current_family :selection
	pal_separator
	@selmode.contribute_palette @palette, 'MHE' if @selmode
	
	#Preview mode
	@palette.set_current_family :cleansing, :algo
	palette_preview_mode

	#Option for calculation of terrain
	@palette.set_current_family :algo
	palette_grid
	palette_hilltop

	#Option for generation of terrain
	@palette.set_current_family :geometry
	palette_option_geometry	
	
	#Abort and Exit
	@palette.set_current_family
	pal_separator
	hidden_proc = proc { @mode == :cleansing && @saved_contours }
	hsh = { :hidden_proc => hidden_proc, :draw_proc => :std_abortexit, :main_color => 'red', :frame_color => 'green', 
	        :tooltip => T6[:T_STR_AbortTool] }
	@palette.declare_button(:pal_abort, hsh) { abort_tool }

	#Rollback
	@palette.set_current_family :cleansing, :algo, :geometry
	ssb = :back
	tip_proc = proc { notify_proc.call :tip, ssb }
	hidden_proc = proc { notify_proc.call :gray, ssb }
	hsh = { :tip_proc => tip_proc, :draw_proc => :rollback, :hidden_proc => hidden_proc }
	@palette.declare_button(:pal_back, hsh) { notify_proc.call :exec, ssb }
	
	#Exit tool
	@palette.set_current_family
	hidden_proc = proc { @mode != :geometry && (@mode != :cleansing || !@saved_contours) }
	hsh = { :hidden_proc => hidden_proc, :draw_proc => :std_exit, :tooltip => T6[:T_STR_ExitTool] }
	@palette.declare_button(:pal_exit, hsh) { exit_tool }
		
	#Execute (or Save options for Cleanser-only mode)
	palette_validate unless @action_code == :cleanser
	palette_cleanser_save

	#Special button for selection	
	@palette.set_current_family :selection
	ssvc = :next_curve
	tip_proc = proc { notify_proc.call :tip, ssvc }
	grayed_proc = proc { notify_proc.call :gray, ssvc }
	hsh = { :tip_proc => tip_proc, :draw_proc => :next_curve, :grayed_proc => grayed_proc }
	@palette.declare_button(:pal_next_curve, hsh) { notify_proc.call :exec, ssvc}
		
	#Additional options
	palette_options_debug
	
	#Creating the floating palettes
	palette_floating_cleansing
	palette_floating_algo
	
	#Associating the palette
	set_palette @palette
end

#PALETTE: Buttons for grid dimensions
def palette_options_debug
	return unless MYDEFPARAM[:TPS_OPTION_Debug]
	@palette.set_current_family :algo
	pal_separator
	
	value_proc = proc { @hsh_options_global[:debug_zone] }
	hsh = { :value_proc => value_proc, :draw_proc => @draw_local, :tooltip => T6[:TIP_DebugOption] }
	@palette.declare_button(:pal_debug_zone, hsh) { @hsh_options_global[:debug_zone] = !@hsh_options_global[:debug_zone] }
end

#PALETTE: Buttons for grid dimensions
def palette_step_label
	@palette.set_current_family
	wid = 0
	[:BOX_Mode_selection, :BOX_Mode_cleansing, :BOX_Mode_algo, :BOX_Mode_geometry].each do |symb|
		text = T6[symb]
		w, = G6.text_size text
		wid = w if w > wid
	end
	text_proc = proc { @mode_text }
	tip_proc = proc { @mode_tip }
	hsh = { :passive => true, :width => wid + 20, :height => 32, :draw_proc => @draw_local, :justif => 'MH', :text_proc => text_proc, :tip_proc => tip_proc }
	@palette.declare_button(:pal_step, hsh)
end

#PALETTE: Buttons for grid dimensions
def palette_cleanser_save
	@palette.set_current_family :cleansing

	txt_save_replace = T6[:BOX_SaveReplace].sub(/\s*&\s*/, "\n& ")
	txt_save_copy = T6[:BOX_SaveCopy].sub(/\s*&\s*/, "\n& ")
	w1, = G6.text_size(txt_save_replace)
	w2, = G6.text_size(txt_save_copy)
	w = [w1, w2].max + 70
	hshb = { :width => w, :height => 32, :justif => 'MH', :bk_color => 'plum', :draw_proc => :std_disk }
	
	hidden_proc = proc { @selmode.empty? }
	hsh = { :hidden_proc => hidden_proc, :text => txt_save_replace, :tooltip => T6[:TIP_SaveReplace] }
	@palette.declare_button(:pal_cleanser_update, hshb, hsh) { save_replace_contours }
	#hsh = { :hidden_proc => hidden_proc, :text => txt_save_copy, :tooltip => T6[:TIP_SaveCopy] }
	#@palette.declare_button(:pal_cleanser_copy, hshb, hsh) { save_copy_contours }
end

#PALETTE: Buttons for grid dimensions
def palette_grid
	hsh = { :passive => true, :draw_proc => :separator_V, :width => 12, :height => 32 }	
	@palette.declare_button :pal_grid_sepa0, hsh
	
	notify_proc = self.method 'notify_from_palette'
	symb_master = :pal_grid
	text = T6[:BOX_Grid_Title]
	wid, = G6.simple_text_size(text)
	wid = [80, wid].max
	tip_proc = proc { notify_proc.call :tip, :grid_nx_ny }
	
	hsh = { :type => 'multi_free', :passive => true, :text => text, :tip_proc => tip_proc,
	        :height => 16, :bk_color => 'lightblue' }
	@palette.declare_button(symb_master, hsh)
	
	tip_proc = proc { notify_proc.call :tip, :grid_lx_ly }
	text_proc = proc { @algo.text_grid_dimension }
	hsh = { :parent => symb_master, :text_proc => text_proc, :width => wid, :tip_proc => tip_proc,
	        :bk_color => 'lightgreen' }
	@palette.declare_button(:pal_grid_nx_ny, hsh) { ask_grid_dimension }		
end

#PALETTE: Buttons for Preview mode
def palette_preview_mode
	hsh = { :passive => true, :draw_proc => :separator_V, :width => 12, :height => 32 }	
	@palette.declare_button :pal_preview_mode_sepa0, hsh

	symb_master = :pal_preview_mode

	hsh = { :type => 'multi_free', :passive => true, :text => T6[:BOX_WorkingView], :bk_color => 'lightblue',
	        :height => 16, :tooltip => T6[:TIP_WorkingView] }
	@palette.declare_button(symb_master, hsh) 

	hshp = { :parent => symb_master, :draw_proc => @draw_local }

	wid = 28	
	value_proc = proc { @mode == :algo && !working_view_active? }
	hsh = { :width => wid, :value_proc => value_proc, :text => "SU", :tooltip => T6[:TIP_PreviewMode_InSitu] }
	@palette.declare_button(:pal_preview_mode_insitu, hshp, hsh) { set_preview_in_situ }	

	hsh = { :parent => symb_master, :passive => true, :draw_proc => :separator_V, :width => 12 }	
	@palette.declare_button :pal_preview_mode_sepa1, hsh
	
	hsh = { :width => wid, :tooltip => T6[:TIP_PreviewMode_Surface] }
	@palette.declare_button(:pal_preview_mode_surface, hshp, hsh) { set_preview_working :surface }	

	hsh = { :width => wid, :tooltip => T6[:TIP_PreviewMode_Map] }
	@palette.declare_button(:pal_preview_mode_map, hshp, hsh) { set_preview_working :map }	

	hsh = { :width => wid, :tooltip => T6[:TIP_PreviewMode_SurfaceMap] }
	@palette.declare_button(:pal_preview_mode_surface_map, hshp, hsh) { set_preview_working :surface_map }	
end

#PALETTE: Buttons for hilltop mode
def palette_hilltop
	hsh = { :passive => true, :draw_proc => :separator_V, :width => 12, :height => 32 }	
	@palette.declare_button :pal_hilltop_sepa0, hsh

	symb_master = :pal_hilltop
	
	value_proc = proc { @algo && @algo.option_hilltop? }
	hsh = { :type => 'multi', :radio => 'true', :passive => true, :text => T6[:BOX_Hilltop], :bk_color => 'lightblue',
	        :height => 16, :default_value => :auto, :tooltip => T6[:TIP_HilltopMain], :value_proc => value_proc }
	@palette.declare_button(symb_master, hsh) { @algo.option_change_hilltop @palette.button_get_value }

	wid = 30	
	hshp = { :parent => symb_master, :draw_proc => @draw_local, :width => wid }

	hsh = { :value => :flat, :tooltip => T6[:TIP_HilltopFlatAll] }
	@palette.declare_button(:pal_hilltop_flat, hshp, hsh)

	hsh = { :value => :auto, :tooltip => T6[:TIP_HilltopAutoAll] }
	@palette.declare_button(:pal_hilltop_auto, hshp, hsh)	
end

#PALETTE: Buttons for hilltop mode
def palette_option_geometry
	hsh = { :passive => true, :draw_proc => :separator_V, :width => 12, :height => 32 }	
	@palette.declare_button :pal_geometry_sepa0, hsh

	txbox = T6[:BOX_OptionGeometry]
	wtot, = G6.simple_text_size(txbox)
	wtot += 40
	
	symb_master = :pal_option_geometry
	
	hsh = { :type => 'multi_free', :passive => true, :text => txbox, :bk_color => 'lightblue',
	        :height => 16, :width => wtot, :tooltip => T6[:TIP_OptionGeometry] }
	@palette.declare_button(symb_master, hsh)

	wid = [wtot / 3, 45].max
	hshp = { :parent => symb_master, :draw_proc => @draw_local, :width => wid }

	value_proc = proc { @hsh_options_global[:geometry_wall] }
	hsh = { :value => :flat, :tooltip => T6[:TIP_OptionGeometryWall], :value_proc => value_proc }
	@palette.declare_button(:pal_option_geometry_wall, hshp, hsh) { @algo.geometry_option_wall_toggle }

	value_proc = proc { @hsh_options_global[:geometry_contour] }
	hsh = { :value => :auto, :tooltip => T6[:TIP_OptionGeometryContour], :value_proc => value_proc }
	@palette.declare_button(:pal_option_geometry_contour, hshp, hsh) { @algo.geometry_option_contour_toggle }

	value_proc = proc { @hsh_options_global[:geometry_map] }
	hsh = { :value => :auto, :tooltip => T6[:TIP_OptionGeometryMap], :value_proc => value_proc }
	@palette.declare_button(:pal_option_geometry_map, hshp, hsh) { @algo.geometry_option_map_toggle }	
end

def palette_validate
	notify_proc = self.method 'notify_from_palette'

	color = Sketchup::Color.new 197, 247, 214
	ssv = :validate
	tip_proc = proc { notify_proc.call :tip, ssv }
	grayed_proc = proc { notify_proc.call :gray, ssv }
	hshc = { :draw_proc => :valid, :height => 32, :hidden_proc => grayed_proc,
             :bk_color => color	}
	
	#Creating the Validate buttons
	[:selection, :cleansing, :algo].each do |symb|
		t6symb = "BOX_Validate_#{symb}".intern
		tipsymb = "TIP_Validate_#{symb}".intern
		palsymb = "pal_#{t6symb}".intern
		lt = T6[t6symb].split '|'
		text = lt[0].strip
		text += "\n" + lt[1].strip if lt[1]
		w, = G6.text_size(text)
		w += 60
		hsh = { :text => text, :width => w, :tooltip => T6[tipsymb] }
		@palette.set_current_family symb	
		pal_separator
		@palette.declare_button(palsymb, hshc, hsh) { notify_proc.call :exec, ssv }
	end
end

def notify_cleanser_ui(code)
	@cleanser.notify_contour_set code
	onMouseMove_zero
end

#PALETTE: Floating palette for Cleansing operations
def palette_floating_cleansing
	@palette.set_current_family :cleansing
	wh = 20
	
	#Declare the floating palette
	flpal = :pal_cleanser_floating
	hshfloat = { :floating => flpal, :row => 0, :width => wh, :height => wh }
	hidden_proc = proc { @mode != :cleansing || !@cleanser || !@cleanser.show_floating_palette? }
	hsh = { :auto_move => [:up], :hidden_proc => hidden_proc }
	@palette.declare_floating flpal, hsh

	#Buttons for removing / inserting contours
	symb = "#{flpal}_accept".intern
	value_proc = proc { @cleanser && @cleanser.contour_all_included? }
	hsh = { :value_proc => value_proc, :draw_proc => :valid_3, :tooltip => T6[:TIP_ContourAccept], :hi_color => 'yellow' }
	@palette.declare_button(symb, hshfloat, hsh) { notify_cleanser_ui :include }

	symb = "#{flpal}_ignore".intern
	value_proc = proc { @cleanser.contour_all_ignored? }
	hsh = { :value_proc => value_proc, :draw_proc => :cross_NE3, :tooltip => T6[:TIP_ContourIgnore], :frame_color => 'red', :hi_color => 'yellow' }
	@palette.declare_button(symb, hshfloat, hsh) { notify_cleanser_ui :ignore }
	
	#Separator
	@palette.declare_separator_floating flpal, { :height => wh }
	
	#Buttons for simplify
	symb = "#{flpal}_simplify".intern
	value_proc = proc { @cleanser.contour_all_simplified? }
	hsh = { :value_proc => value_proc, :draw_proc => @draw_local, :tooltip => T6[:MNU_SimplifyContour], :hi_color => 'yellow' }
	@palette.declare_button(symb, hshfloat, hsh) { notify_cleanser_ui :simplify }

	symb = "#{flpal}_simplify_no".intern
	value_proc = proc { @cleanser.contour_all_not_simplified? }
	hsh = { :value_proc => value_proc, :draw_proc => @draw_local, :tooltip => T6[:MNU_UnSimplifyContour], :hi_color => 'yellow' }
	@palette.declare_button(symb, hshfloat, hsh) { notify_cleanser_ui :unsimplify }
	
end

#PALETTE: Floating palette for Calculation operations (algo)
def palette_floating_algo
	@palette.set_current_family :algo
	wh = 32
	
	#Declare the floating palette
	flpal = :pal_algo_floating
	hshfloat = { :floating => flpal, :row => 0 }
	hidden_proc = proc { @mode != :algo || !@algo || !@algo.show_floating_palette? }
	hsh = { :auto_move => [:up], :hidden_proc => hidden_proc }
	@palette.declare_floating flpal, hsh
	
	#Hilltop management
	symb_master = "#{flpal}_hilltop_title".intern
	
	hidden_proc = proc { !@algo.option_hilltop_possible_on_selected? }
	value_proc = proc { @algo.option_hilltop_value_on_selected }
	hsh = { :type => 'multi', :radio => 'true', :passive => true, :text => T6[:BOX_Hilltop], :bk_color => 'lightblue',
	        :height => 16, :default_value => :auto, :tooltip => T6[:TIP_HilltopMain], :value_proc => value_proc,
			:hidden_proc => hidden_proc }
	@palette.declare_button(symb_master, hshfloat, hsh) { @algo.notify_contour_set :hilltop, @palette.button_get_value, @algo.contour_selection }

	wid = 30	
	hshp = { :parent => symb_master, :draw_proc => @draw_local, :width => wid }

	symb = "#{flpal}_hilltop_flat".intern
	hsh = { :value => :flat, :tooltip => T6[:TIP_HilltopFlat] }
	@palette.declare_button(symb, hshfloat, hshp, hsh)	

	symb = "#{flpal}_hilltop_auto".intern
	hsh = { :value => :auto, :tooltip => T6[:TIP_HilltopAuto] }
	@palette.declare_button(symb, hshfloat, hshp, hsh)	
end

def palette_floating_cliff
	#Cliff management
	wid = 24	
	widf = 80
	#widtot = 2 * wid + widf
	symb_master = "#{flpal}_cliff_title".intern
	
	hsh = { :type => 'multi_free', :passive => true, :text => T6[:BOX_Cliff], :bk_color => 'lightblue',
	        :height => 16, :default_value => 0, :tooltip => T6[:TIP_CliffMain] }
	#@palette.declare_button(symb_master, hshfloat, hsh)

	hshp = { :parent => symb_master, :height => 16 }

	symb = "#{flpal}_no_cliff".intern
	hsh = { :draw_proc => @draw_local, :width => wid, :tooltip => T6[:TIP_CliffNO] }
	@palette.declare_button(symb, hshfloat, hshp, hsh) { @algo.notify_contour_set :cliff, 0, @algo.contour_selection }

	#Input field
	hshi = { :vtype => :integer, :vmin => 0, :vmax => 100, :vincr => 10, 
	         :vsprintf => "%d%", :vprompt => T6[:TIP_CliffVal] }
	get_proc = proc { 50 }
	set_proc = proc { |val| @algo.notify_contour_set :cliff, val, @algo.contour_selection }
	input = Traductor::InputField.new hshi, { :get_proc => get_proc, :set_proc => set_proc }
	
	symb = "#{flpal}_cliff_val".intern
	hsh = { :bk_color => 'yellow', :input => input, :width => widf, :tooltip => T6[:TIP_CliffVal] }
	@palette.declare_button symb, hshfloat, hshp, hsh 

	symb = "#{flpal}_true_cliff".intern
	hsh = { :draw_proc => @draw_local, :width => wid, :tooltip => T6[:TIP_CliffTrue] }
	@palette.declare_button(symb, hshfloat, hshp, hsh) { @algo.notify_contour_set :cliff, 100, @algo.contour_selection }	
	
	#Separator
	@palette.declare_separator_floating flpal, { :height => wh }
end

#PALETTE: Drawing the Blason
def draw_button_blason(symb, dx, dy)
	lst_gl = []
	dx2 = dx/2
	dy2 = dy/2
	dym = dy * 3 / 4
	dyb = dym-4
	ptmid = Geom::Point3d.new dx2, dy2
	tr = Geom::Transformation.rotation ptmid, Z_AXIS, 45.degrees
	
	lip = [[[6, 3], [dx2, 3]], [[1, 8], [8, 6], [dx2, 4]], [[1, 14], [7, 11], [12, 8], [dx2, 6]], [[1, 20], [7, 16], [12, 10], [dx2, 9]],
	       [[4, 25], [7, 20], [16, 18], [dx2+5, 13]], [[dx2, dy-3], [dx2+10, 20]]]
	
	lip.each do |ls|
		pts = ls.collect { |a| Geom::Point3d.new *a }
		lst_gl.push [GL_LINE_STRIP, pts, 'blue', 2, '']
	end	
		
	pts = G6.pts_square 22, 6, 4
	lst_gl.push [GL_LINE_LOOP, pts, 'red', 2, '']
	
	lst_gl
end

#--------------------------------------------------------------
# Custom drawing for palette buttons
#--------------------------------------------------------------

#Custom drawing of buttons
def draw_button_opengl(symb, dx, dy, main_color, frame_color, selected)
	code = symb.to_s
	lst_gl = []
	dx2 = dx / 2
	dy2 = dy / 2
	grayed = @palette.button_is_grayed?(symb)
	color = (grayed) ? 'gray' : frame_color
	
	case code

	#Preview Modes
	when /step/
		rect = G6.pts_round_rectangle(-1, -1, dx+1, dy+1, 6)
		lst_gl.push [GL_POLYGON, rect, 'gold']
		lst_gl.push [GL_LINE_LOOP, rect, 'goldenrod', 2, '']

	when /simplify/
		dec = 2
		pts = [Geom::Point3d.new(dx-dec, dy-dec-1), Geom::Point3d.new(dec, dy2), Geom::Point3d.new(dx-dec, dec+1)]
		lst_gl.push [GL_LINE_STRIP, pts, 'green', 3, '']
		if code =~ /_no\Z/
			pts = [Geom::Point3d.new(dx-dec-3, 1), Geom::Point3d.new(dec, dy)]
			lst_gl.push [GL_LINE_STRIP, pts, 'red', 3, '']
		end
		
	#Preview Modes
	when /preview_mode_(.*)/
		sub_code = $1
		map_only = (sub_code =~ /\Amap/)
		surface_only = (sub_code =~ /surface\Z/)
		if sub_code =~ /map/
			lx = (map_only) ? [0, 0.33, 0.66, 1] : [0.66, 1]
			pts = [[], [], [], []]
			lx.each_with_index do |x, ix|
				[0, 0.5, 1].each_with_index do |y, iy|
					pts[ix][iy] = Geom::Point3d.new dx * x, dy * y, 0
				end	
			end
			for ix in 0..lx.length-2
				for iy in 0..1
					color = ((ix + iy).modulo(2) == 0) ? 'lightyellow' : 'lightgreen'
					rect = [pts[ix][iy], pts[ix+1][iy], pts[ix+1][iy+1], pts[ix][iy+1]]
					lst_gl.push [GL_POLYGON, rect, color]
					lst_gl.push [GL_LINE_LOOP, rect, 'gray', 1, '']
				end
			end	
		end
		if sub_code =~ /surface/
			w = (surface_only) ? dx : dx2
			origin = Geom::Point3d.new(1,0)
			axe = [origin, Geom::Point3d.new(w, 0)]
			lst_gl.push [GL_LINE_STRIP, axe, 'red', 2, '']
			axe = [origin, Geom::Point3d.new(1, dy)]
			lst_gl.push [GL_LINE_STRIP, axe, 'blue', 2, '']
			axe = [origin, Geom::Point3d.new(dx2, dy2)]
			lst_gl.push [GL_LINE_STRIP, axe, 'green', 2, '']
		end
		
		if sub_code =~ /situ/i
			rect = G6.pts_rectangle 0, 0, dx, dy
			lst_gl.push [GL_POLYGON, rect, (selected) ? 'yellow' : 'lightgrey']
			lst_gl.push [GL_LINE_LOOP, rect, 'blue', 1, '']
		end	
	
	#Hilltop handling
	when /hilltop_(.*)/
		pt1 = Geom::Point3d.new(1, 1)
		pt2 = Geom::Point3d.new(dx/4, dy2)
		pt3 = Geom::Point3d.new(dx - dx/4, dy2)
		pt4 = Geom::Point3d.new(dx-1, 1)
		if $1 == 'flat'
			pts = [pt2, pt3]
		else
			ptt0 = Geom::Point3d.new(dx2-dx/4, dy-dy/4)
			ptt1 = Geom::Point3d.new(dx2, dy-1)
			ptt2 = Geom::Point3d.new(dx2+dx/4, dy-dy/4)
			pts = [pt2, ptt0, ptt1, ptt2, pt3]
		end	
		lst_gl.push [GL_LINES, [pt1, pt2, pt3, pt4], 'gray', 2, '']
		lst_gl.push [GL_LINE_STRIP, pts, 'blue', 2, '']
	
	when /option_geometry_wall/i
		lpi = [[dx, dy-8], [dx-5, dy-3], [dx2, dy], [5, dy-1], [1, dy-5]]
		lpti = lpi.collect { |a| Geom::Point3d.new *a }
		pts = [[1, 1], [dx, 1]].collect { |a| Geom::Point3d.new *a } + lpti
		lst_gl.push [GL_POLYGON, pts, 'khaki']
		lst_gl.push [GL_LINE_LOOP, pts, 'gray', 1, '']
		pts = []
		lpti.each_with_index { |pt, i| pts.push pt, Geom::Point3d.new(lpti[i].x, 1) }
		lst_gl.push [GL_LINES, pts, 'gray', 1, '']
		
	when /option_geometry_contour/i
		lpi = [[1, 1], [dx/4, dy-2], [dx2, 8], [3*dx/4, dy-2]]
		lst_gl.push [GL_LINE_STRIP, lpi.collect { |a| Geom::Point3d.new *a }, 'blue', 2, '']
		lpi = [[dx/4, 1], [dx2, dy2], [3*dx/4, dy2-3], [dx, dy2+3]]
		lst_gl.push [GL_LINE_STRIP, lpi.collect { |a| Geom::Point3d.new *a }, 'blue', 2, '']
		
	when /option_geometry_map/i
		dx4 = dx / 4
		dy4 = dy / 4
		lpi = [[0, 0], [dx, 0], [dx, dy], [0, dy]]
		lst_gl.push [GL_POLYGON, lpi.collect { |a| Geom::Point3d.new *a }, 'moccasin']
		lst_gl.push [GL_LINE_LOOP, lpi.collect { |a| Geom::Point3d.new *a }, 'black', 1, '']
		
		lip = [[[1, dy2], [dx4, 1]], [[dx4, dy-1], [dx2, 1]], [[dx2, dy-1], [3*dx4, 1]], [[3*dx4, dy-1], [dx, dy2]]]
		lip.each do |ls|
			pts = ls.collect { |a| Geom::Point3d.new *a }
			lst_gl.push [GL_LINE_STRIP, pts, 'black', 2, '']
		end	
		
	when /debug_zone/i	
		lpi = [[1, 8], [dx/4, dy-2], [dx2, dy2+6], [3*dx/4, dy-2]]
		lst_gl.push [GL_LINE_STRIP, lpi.collect { |a| Geom::Point3d.new *a }, 'red', 2, '']
		lpi = [[dx/4, 1], [dx2, dy2-3], [3*dx/4, dy2-6], [dx, dy2+3]]
		lst_gl.push [GL_LINE_STRIP, lpi.collect { |a| Geom::Point3d.new *a }, 'red', 2, '']
		pts = G6.pts_square dx2, dy2, 3
		lst_gl.push [GL_POLYGON, pts, 'blue']
		
	end	#case code
	
	lst_gl
end

#---------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------
# GEOMETRY: Mouse Movement Methods
#---------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------

#GEOMETRY: Execute the generation of Terrain geometry
def geometry_execute
	@geometry_processing = true
	restore_view
	@suops.abort_operation
	
	begin
		@algo.geometry_execute @suops, attribute_group_retrieve()
	rescue Exception => e
		rescue_errors e, "GENERATION OF TERRAIN"
	end
end

#GEOMETRY: Rollback the generation
def geometry_rollback
	case @mode
	when :algo
		restore_view
		@suops.abort_restart_operation
		set_state_mode :cleansing
	when :geometry
		@suops.abort_restart_operation
		set_state_mode :algo
		onMouseMove_zero
	end	
end

#GEOMETRY: Notification of the Termination of the geometry
def geometry_terminate(time)
	@geometry_processing = false
	@algo.geometry_terminate
	if time < 0
		set_state_mode :algo
	else
		set_state_mode :geometry
	end
	onMouseMove_zero
	onSetCursor
end

#---------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------
# VCB: Handling VCB inputs
#---------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------
	
def enableVCB? ; @mode == :algo ; end
	
def onUserText(text, view)
	handle_VCB(text, view) if @mode == :algo
end	
	
#VCB: Declare vcb formats
def init_VCB	
	@vcb = Traductor::VCB.new
	@vcb.declare_input_format :nx, "i"
	@vcb.declare_input_format :dx, "l"
end

#VCB: Handle VCB input
def handle_VCB(text, view)
	return UI.beep unless @vcb
	nb_errors = @vcb.process_parsing(text)
	
	if nb_errors > 0
		lmsg = []
		@vcb.each_error { |symb, s| lmsg.push "<#{s}>" }
		msg = "#{T6[:T_ERROR_InVCB]} \"#{text}\"  --> #{lmsg.join(' - ')}"
		@palette.set_message msg, 'E'
		view.invalidate
	else
		action_from_VCB(text)
	end	
end

#VCB: Execute actions from the VCB inputs for the grid dimensions	
def action_from_VCB(text)
	lvalue = []
	@vcb.each_result do |symb, val, suffix|
		lvalue.push [symb, val]
	end
	
	@algo.grid_check_dimension_as_VCB(lvalue, text)	
end

#---------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------
# VISU: Management of Insitu and Working views
#---------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------

#VISU: View requested is in_situ
def set_preview_in_situ
	return unless working_view_active?
	@camera_last_working = G6.camera_duplicate @view.camera
	@preview_in_situ = true
	@view.camera = @camera_last_in_situ, 0
	change_view :in_situ
end

#VISU: View requested is a working view
def set_preview_working(symb=nil)
	in_situ = !working_view_active?
	if @camera_last_in_situ
		@camera_last_in_situ = G6.camera_duplicate(@view.camera) unless working_view_active?
	else	
		@camera_original = G6.camera_duplicate(@view.camera) unless working_view_active?
		@camera_last_in_situ = (MYDEFPARAM[:TPS_OPTION_ZoomInSitu]) ? @myclass.camera(:geometry) : @camera_original
	end	
	@camera_last_in_situ.description = ""
	@preview_in_situ = false
	change_view :working
	@view.camera = ((!symb && @camera_last_working) ? @camera_last_working : @myclass.camera(symb)), 0
	@myclass.dessin_refresh_when_view_changed
	update_dessin_on_resume
end

#VISU: Swicth to the working view if necessary
def switch_view
	return if working_view_active? || @preview_in_situ
	set_preview_working
end

#VISU: Restore the full SU View
def restore_view
	return unless @myclass
	change_view :restore
	if working_view_active?	
		@camera_last_working = G6.camera_duplicate @view.camera
		@view.camera = (@aborting) ? @camera_original : @camera_last_in_situ
	end
end

#VISU: Check if the current view is a working view
def working_view_active?
	@view.camera.description =~ /TOPOSHAPER_/	
end

def change_view(view_state)
	@temp_layer = Traductor::TemporaryLayer.new unless @temp_layer
	if view_state == :restore || view_state == :in_situ
		@temp_layer.restore
	else
		@temp_layer.switch_to_temp_layer
	end
	@myclass.edgewire_set_layer @temp_layer.layer
	@myclass.notify_working_view view_state
end

def compte_camera_in_situ

end

end	#class TopoShaperTool

end	#End Module TopoShaper
