=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed July 2010 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   bootstrap_TopoShaper.rb
# Original Date	:   11 July 2010 - version 1.0
# Description	:   Loader for the TopoShaper scripts
# Usage			:   See Tutorial and Quick Ref Card in PDF format
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module TopoShaper

#---------------------------------------------------------------------------------------------------------------------------
# Overall Configuration for the Tools	
#---------------------------------------------------------------------------------------------------------------------------

MENU_PERSO = [:T_MenuTools_Fredo6Collection, "Tools", :T_MenuPlugins_Fredo6Collection, "Plugins"]

#---------------------------------------------------------------------------------------------------------------------------
#Constants for TOS  Modules (do not translate here, use Translation Dialog Box instead)	
#---------------------------------------------------------------------------------------------------------------------------

#Menu and Toolbar icons
T6[:TPS_QUADTERRAIN_Menu] = "Generate QuadMesh Terrain"
T6[:TPS_QUADTERRAIN_Tooltip] = "Generate terrain from iso-contours with Quad mesh"
TPS_QUADTERRAIN_Icon = "QuadTerrain"
T6[:TPS_EDITTERRAIN_Menu] = "Edit Terrain"
T6[:TPS_EDITTERRAIN_Tooltip] = "Edit a previously generated terrain"

#--------------------------------------------------------
# Default Parameters
#--------------------------------------------------------

T6[:TPS_DEFAULT_SECTION_General] = "Parameters at startup of Sketchup"
T6[:TPS_OPTION_Debug] = "Debug option to show algorithm details"
T6[:TPS_OPTION_ZoomInSitu] = "Automatic Zooming on generated terrain"

#---------------------------------------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------------------------------------
# EXTERNAL API to launch RoundCorner menus
# Action code must be passed as a symbol
#---------------------------------------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------------------------------------
	
def TopoShaper.launch_action(action_code)
	MYPLUGIN.launch_action action_code
end

#--------------------------------------------------------------
# Condition for contextual menu selection
#--------------------------------------------------------------

def TopoShaper.selection_contextual?
	selection = Sketchup.active_model.selection
	return nil if selection.empty?
	
	#Selection of an existing terrain group
	lst_groups = selection.grep(Sketchup::Group)
	lst_groups.each do |group|
		return :edit if group.attribute_dictionary("TOPOSHAPER", false)
	end
	
	#Selection if a group or a component instance with edges
	lst_groups.each { |group| return :edge if group.entities.find { |e| e.instance_of?(Sketchup::Edge) } }
	lst_comps = selection.grep(Sketchup::ComponentInstance)
	lst_comps.each { |comp| return :edge if comp.definition.entities.find { |e| e.instance_of?(Sketchup::Edge) } }
	nil
end

#--------------------------------------------------------
# Startup routine (called from LibFredo6)
#--------------------------------------------------------
	
#Create all commands with menus and toolbar icons
def TopoShaper.startup	

	#Conditions for contextual menus
	tc_edit = Traductor::TestCondition.new() { TopoShaper.selection_contextual? == :edit }
	tc_edge = Traductor::TestCondition.new() { TopoShaper.selection_contextual? == :edge }
	submenu = "TopoShaper"
	
	#Top menu
	MYPLUGIN.declare_topmenu nil, MENU_PERSO

	#Declaring command icons
	MYPLUGIN.declare_separator
	MYPLUGIN.declare_command(:TPS_QUADTERRAIN_) { TopoShaper.launch_action :quad_terrain }
	MYPLUGIN.declare_context_handler(:TPS_EDITTERRAIN_, tc_edit, nil, submenu) { TopoShaper.launch_action :quad_terrain }
	MYPLUGIN.declare_context_handler(:TPS_QUADTERRAIN_, tc_edge, nil, submenu) { TopoShaper.launch_action :quad_terrain }
	
	#Declaring the default parameter
	TopoShaper.default_param
	
	#Startup of the Plugin
	MYPLUGIN.go	
end
	
#Declaring the Default Parameters	
def TopoShaper.default_param
	dp = MYDEFPARAM
	dp.separator :TPS_DEFAULT_SECTION_General
	dp.declare :TPS_OPTION_Debug, false, "B"	
	dp.declare :TPS_OPTION_ZoomInSitu, true, "B"	
end
		
end	#End Module TopoShaper
